"""RDS tests covering integrations with other service backends."""

import json
import uuid

import boto3
import pytest
from botocore import waiter, xform_name
from botocore.exceptions import ClientError

from tests import aws_verified

from . import DEFAULT_REGION


@aws_verified
@pytest.mark.aws_verified
def test_db_cluster_managed_master_user_password_lifecycle():
    # Clients
    kms = boto3.client("kms", region_name=DEFAULT_REGION)
    rds = boto3.client("rds", region_name=DEFAULT_REGION)
    secretsmanager = boto3.client("secretsmanager", region_name=DEFAULT_REGION)

    # Waiters
    db_cluster_modifications_complete = get_custom_waiter(
        "db_cluster_modifications_complete", rds
    )
    db_cluster_available = rds.get_waiter("db_cluster_available")
    db_cluster_deleted = rds.get_waiter("db_cluster_deleted")
    master_user_secret_available = get_custom_waiter(
        "db_cluster_master_user_secret_available", rds
    )
    master_user_secret_deleted = get_custom_waiter("secret_deleted", secretsmanager)

    # KMS keys
    resp = kms.describe_key(KeyId="alias/aws/secretsmanager")
    default_kms_key_arn = resp["KeyMetadata"]["Arn"]
    resp = kms.create_key(KeyUsage="ENCRYPT_DECRYPT", KeySpec="SYMMETRIC_DEFAULT")
    custom_kms_key_arn = resp["KeyMetadata"]["Arn"]
    custom_kms_key_id = resp["KeyMetadata"]["KeyId"]

    # Create DBCluster with managed master user password
    db_cluster_identifier = "test-" + str(uuid.uuid4())
    resp = rds.create_db_cluster(
        DBClusterIdentifier=db_cluster_identifier,
        Engine="aurora-postgresql",
        MasterUsername="root",
        ManageMasterUserPassword=True,
    )
    db_cluster = resp["DBCluster"]
    master_user_secret = db_cluster["MasterUserSecret"]
    assert master_user_secret["KmsKeyId"] == default_kms_key_arn
    assert master_user_secret["SecretStatus"] == "creating"
    # Check the secret in SecretsManager
    secret_arn = master_user_secret["SecretArn"]
    master_user_secret_available.wait(DBClusterIdentifier=db_cluster_identifier)
    secret = secretsmanager.describe_secret(SecretId=secret_arn)
    assert str(secret["Name"]).startswith("rds!cluster")
    assert secret["OwningService"] == "rds"
    # Check that RDS sees the secret as active
    db_cluster_available.wait(DBClusterIdentifier=db_cluster_identifier)
    resp = rds.describe_db_clusters(DBClusterIdentifier=db_cluster_identifier)
    db_cluster = resp["DBClusters"][0]
    master_user_secret = db_cluster["MasterUserSecret"]
    assert master_user_secret["SecretStatus"] == "active"
    # Check that managed secret has correct structure
    resp = secretsmanager.get_secret_value(SecretId=secret_arn)
    secret = json.loads(resp["SecretString"])
    assert "username" in secret
    assert "password" in secret

    # Disable password management
    resp = rds.modify_db_cluster(
        DBClusterIdentifier=db_cluster_identifier,
        ManageMasterUserPassword=False,
        MasterUserPassword="Non-Managed-Pa55word!",
        ApplyImmediately=True,
    )
    db_cluster = resp["DBCluster"]
    # TODO: This passes on AWS but is not implemented yet in moto
    # assert "MasterUserPassword" in db_cluster["PendingModifiedValues"]
    db_cluster_modifications_complete.wait(DBClusterIdentifier=db_cluster_identifier)
    resp = rds.describe_db_clusters(DBClusterIdentifier=db_cluster_identifier)
    db_cluster = resp["DBClusters"][0]
    assert "MasterUserSecret" not in db_cluster
    # Secret should be deleted
    master_user_secret_deleted.wait(SecretId=secret_arn)
    with pytest.raises(ClientError) as exc:
        secretsmanager.describe_secret(SecretId=secret_arn)
    assert exc.value.response["Error"]["Code"] == "ResourceNotFoundException"

    # Re-enable password management, this time with a custom KMS key
    resp = rds.modify_db_cluster(
        DBClusterIdentifier=db_cluster_identifier,
        ManageMasterUserPassword=True,
        MasterUserSecretKmsKeyId=custom_kms_key_arn,
        ApplyImmediately=True,
    )
    db_cluster = resp["DBCluster"]
    # TODO: This passes on AWS but is not implemented yet in moto
    # assert "MasterUserPassword" in db_cluster["PendingModifiedValues"]
    master_user_secret = db_cluster["MasterUserSecret"]
    assert master_user_secret["KmsKeyId"] == custom_kms_key_arn
    assert master_user_secret["SecretStatus"] == "creating"
    # Check the secret in SecretsManager
    secret_arn = master_user_secret["SecretArn"]
    master_user_secret_available.wait(DBClusterIdentifier=db_cluster_identifier)
    secret = secretsmanager.describe_secret(SecretId=secret_arn)
    assert str(secret["Name"]).startswith("rds!cluster")
    assert secret["OwningService"] == "rds"
    # Check that RDS sees the secret as active
    db_cluster_modifications_complete.wait(DBClusterIdentifier=db_cluster_identifier)
    resp = rds.describe_db_clusters(DBClusterIdentifier=db_cluster_identifier)
    db_cluster = resp["DBClusters"][0]
    master_user_secret = db_cluster["MasterUserSecret"]
    assert master_user_secret["SecretStatus"] == "active"

    # Rotate managed password
    resp = rds.modify_db_cluster(
        DBClusterIdentifier=db_cluster_identifier,
        RotateMasterUserPassword=True,
        ApplyImmediately=True,
    )
    db_cluster = resp["DBCluster"]
    master_user_secret = db_cluster["MasterUserSecret"]
    assert master_user_secret["SecretStatus"] == "rotating"
    # Check that RDS sees the secret as active after rotation
    master_user_secret_available.wait(DBClusterIdentifier=db_cluster_identifier)
    resp = rds.describe_db_clusters(DBClusterIdentifier=db_cluster_identifier)
    db_cluster = resp["DBClusters"][0]
    master_user_secret = db_cluster["MasterUserSecret"]
    assert master_user_secret["SecretStatus"] == "active"

    # Delete the cluster
    resp = rds.delete_db_cluster(
        DBClusterIdentifier=db_cluster_identifier,
        SkipFinalSnapshot=True,
        DeleteAutomatedBackups=True,
    )
    db_cluster = resp["DBCluster"]
    assert "MasterUserSecret" in db_cluster
    # Deleting the cluster should delete the secret
    db_cluster_deleted.wait(DBClusterIdentifier=db_cluster_identifier)
    master_user_secret_deleted.wait(SecretId=secret_arn)
    with pytest.raises(ClientError) as exc:
        secretsmanager.describe_secret(SecretId=secret_arn)
    assert exc.value.response["Error"]["Code"] == "ResourceNotFoundException"
    # Delete the custom KMS key we created
    kms.schedule_key_deletion(KeyId=custom_kms_key_id, PendingWindowInDays=7)


@aws_verified
@pytest.mark.aws_verified
def test_db_instance_managed_master_user_password_lifecycle():
    # Clients
    kms = boto3.client("kms", region_name=DEFAULT_REGION)
    rds = boto3.client("rds", region_name=DEFAULT_REGION)
    secretsmanager = boto3.client("secretsmanager", region_name=DEFAULT_REGION)

    # Waiters
    db_instance_modifications_complete = get_custom_waiter(
        "db_instance_modifications_complete", rds
    )
    db_instance_available = rds.get_waiter("db_instance_available")
    db_instance_deleted = rds.get_waiter("db_instance_deleted")
    master_user_secret_available = get_custom_waiter(
        "db_instance_master_user_secret_available", rds
    )
    master_user_secret_deleted = get_custom_waiter("secret_deleted", secretsmanager)

    # KMS keys
    resp = kms.describe_key(KeyId="alias/aws/secretsmanager")
    default_kms_key_arn = resp["KeyMetadata"]["Arn"]
    resp = kms.create_key(KeyUsage="ENCRYPT_DECRYPT", KeySpec="SYMMETRIC_DEFAULT")
    custom_kms_key_arn = resp["KeyMetadata"]["Arn"]
    custom_kms_key_id = resp["KeyMetadata"]["KeyId"]

    # Create DBInstance with managed master user password
    db_instance_identifier = "test-" + str(uuid.uuid4())
    resp = rds.create_db_instance(
        AllocatedStorage=200,
        DBInstanceIdentifier=db_instance_identifier,
        DBInstanceClass="db.m7g.large",
        Engine="postgres",
        MasterUsername="root",
        ManageMasterUserPassword=True,
    )
    db_instance = resp["DBInstance"]
    master_user_secret = db_instance["MasterUserSecret"]
    assert master_user_secret["KmsKeyId"] == default_kms_key_arn
    assert master_user_secret["SecretStatus"] == "creating"
    # Check the secret in SecretsManager
    secret_arn = master_user_secret["SecretArn"]
    master_user_secret_available.wait(DBInstanceIdentifier=db_instance_identifier)
    secret = secretsmanager.describe_secret(SecretId=secret_arn)
    assert str(secret["Name"]).startswith("rds!db")
    assert secret["OwningService"] == "rds"
    # Check that RDS sees the secret as active
    db_instance_available.wait(DBInstanceIdentifier=db_instance_identifier)
    resp = rds.describe_db_instances(DBInstanceIdentifier=db_instance_identifier)
    db_instance = resp["DBInstances"][0]
    master_user_secret = db_instance["MasterUserSecret"]
    assert master_user_secret["SecretStatus"] == "active"
    # Check that managed secret has correct structure
    resp = secretsmanager.get_secret_value(SecretId=secret_arn)
    secret = json.loads(resp["SecretString"])
    assert "username" in secret
    assert "password" in secret

    # Disable password management
    resp = rds.modify_db_instance(
        DBInstanceIdentifier=db_instance_identifier,
        ManageMasterUserPassword=False,
        MasterUserPassword="Non-Managed-Pa55word!",
        ApplyImmediately=True,
    )
    db_instance = resp["DBInstance"]
    # TODO: This passes on AWS but is not implemented yet in moto
    # assert "MasterUserPassword" in db_cluster["PendingModifiedValues"]
    db_instance_modifications_complete.wait(DBInstanceIdentifier=db_instance_identifier)
    resp = rds.describe_db_instances(DBInstanceIdentifier=db_instance_identifier)
    db_instance = resp["DBInstances"][0]
    assert "MasterUserSecret" not in db_instance
    # Secret should be deleted
    master_user_secret_deleted.wait(SecretId=secret_arn)
    with pytest.raises(ClientError) as exc:
        secretsmanager.describe_secret(SecretId=secret_arn)
    assert exc.value.response["Error"]["Code"] == "ResourceNotFoundException"

    # Re-enable password management, this time with a custom KMS key
    resp = rds.modify_db_instance(
        DBInstanceIdentifier=db_instance_identifier,
        ManageMasterUserPassword=True,
        MasterUserSecretKmsKeyId=custom_kms_key_arn,
        ApplyImmediately=True,
    )
    db_instance = resp["DBInstance"]
    # TODO: This passes on AWS but not implemented yet in moto
    # assert "MasterUserPassword" in db_cluster["PendingModifiedValues"]
    master_user_secret = db_instance["MasterUserSecret"]
    assert master_user_secret["KmsKeyId"] == custom_kms_key_arn
    assert master_user_secret["SecretStatus"] == "creating"
    # Check the secret in SecretsManager
    secret_arn = master_user_secret["SecretArn"]
    master_user_secret_available.wait(DBInstanceIdentifier=db_instance_identifier)
    secret = secretsmanager.describe_secret(SecretId=secret_arn)
    assert str(secret["Name"]).startswith("rds!db")
    assert secret["OwningService"] == "rds"
    # Check that RDS sees the secret as active
    db_instance_modifications_complete.wait(DBInstanceIdentifier=db_instance_identifier)
    resp = rds.describe_db_instances(DBInstanceIdentifier=db_instance_identifier)
    db_instance = resp["DBInstances"][0]
    master_user_secret = db_instance["MasterUserSecret"]
    assert master_user_secret["SecretStatus"] == "active"

    # Rotate managed password
    resp = rds.modify_db_instance(
        DBInstanceIdentifier=db_instance_identifier,
        RotateMasterUserPassword=True,
        ApplyImmediately=True,
    )
    db_instance = resp["DBInstance"]
    master_user_secret = db_instance["MasterUserSecret"]
    assert master_user_secret["SecretStatus"] == "rotating"
    # Check that RDS sees the secret as active after rotation
    master_user_secret_available.wait(DBInstanceIdentifier=db_instance_identifier)
    resp = rds.describe_db_instances(DBInstanceIdentifier=db_instance_identifier)
    db_instance = resp["DBInstances"][0]
    master_user_secret = db_instance["MasterUserSecret"]
    assert master_user_secret["SecretStatus"] == "active"

    # Delete the instance
    resp = rds.delete_db_instance(
        DBInstanceIdentifier=db_instance_identifier,
        SkipFinalSnapshot=True,
        DeleteAutomatedBackups=True,
    )
    db_instance = resp["DBInstance"]
    assert "MasterUserSecret" in db_instance
    # Deleting the instance should delete the secret
    db_instance_deleted.wait(DBInstanceIdentifier=db_instance_identifier)
    master_user_secret_deleted.wait(SecretId=secret_arn)
    with pytest.raises(ClientError) as exc:
        secretsmanager.describe_secret(SecretId=secret_arn)
    assert exc.value.response["Error"]["Code"] == "ResourceNotFoundException"
    # Delete the custom KMS key we created
    kms.schedule_key_deletion(KeyId=custom_kms_key_id, PendingWindowInDays=7)


def get_custom_waiter(waiter_name, client):
    """Some useful waiters that are not present in botocore."""
    config = {
        "version": 2,
        "waiters": {
            "DBClusterModificationsComplete": {
                "delay": 15,
                "operation": "DescribeDBClusters",
                "maxAttempts": 60,
                "acceptors": [
                    {
                        "state": "success",
                        "matcher": "path",
                        "argument": "contains(keys(DBClusters[0]), 'PendingModifiedValues') && length(DBClusters[0].PendingModifiedValues) == `0`",
                        "expected": True,
                    },
                    {
                        "state": "success",
                        "matcher": "path",
                        "argument": "contains(keys(DBClusters[0]), 'PendingModifiedValues')",
                        "expected": False,
                    },
                ],
            },
            "DBClusterMasterUserSecretAvailable": {
                "delay": 15,
                "operation": "DescribeDBClusters",
                "maxAttempts": 60,
                "acceptors": [
                    {
                        "expected": "active",
                        "matcher": "pathAll",
                        "state": "success",
                        "argument": "DBClusters[].MasterUserSecret.SecretStatus",
                    },
                ],
            },
            "DBInstanceModificationsComplete": {
                "delay": 15,
                "operation": "DescribeDBInstances",
                "maxAttempts": 60,
                "acceptors": [
                    {
                        "state": "success",
                        "matcher": "path",
                        "argument": "contains(keys(DBInstances[0]), 'PendingModifiedValues') && length(DBInstances[0].PendingModifiedValues) == `0`",
                        "expected": True,
                    },
                    {
                        "state": "success",
                        "matcher": "path",
                        "argument": "contains(keys(DBInstances[0]), 'PendingModifiedValues')",
                        "expected": False,
                    },
                ],
            },
            "DBInstanceMasterUserSecretAvailable": {
                "delay": 15,
                "operation": "DescribeDBInstances",
                "maxAttempts": 60,
                "acceptors": [
                    {
                        "expected": "active",
                        "matcher": "pathAll",
                        "state": "success",
                        "argument": "DBInstances[].MasterUserSecret.SecretStatus",
                    },
                ],
            },
            "SecretDeleted": {
                "delay": 15,
                "operation": "DescribeSecret",
                "maxAttempts": 60,
                "acceptors": [
                    {
                        "expected": "ResourceNotFoundException",
                        "matcher": "error",
                        "state": "success",
                    },
                ],
            },
        },
    }
    model = waiter.WaiterModel(config)
    mapping = {}
    for name in model.waiter_names:
        mapping[xform_name(name)] = name
    return waiter.create_waiter_with_client(mapping[waiter_name], model, client)



from __future__ import annotations

import copy
import datetime
import re
from collections import OrderedDict
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple

from botocore.utils import merge_dicts

SECONDS_IN_ONE_DAY = 24 * 60 * 60


@dataclass
class FilterDef:
    # A list of object attributes to check against the filter values.
    # Set to None if filter is not yet implemented in `moto`.
    attrs_to_check: list[str] | None
    # Description of the filter, e.g. 'Object Identifiers'.
    # Used in filter error messaging.
    description: str
    # Make comparison case-insensitive.
    case_insensitive: bool = False


class DbInstanceEngine(str, Enum):
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/rds/client/create_db_instance.html
    # 2023-11-08
    AURORA_MYSQL = "aurora-mysql"
    AURORA_POSTGRESQL = "aurora-postgresql"
    CUSTOM_ORACLE_EE = "custom-oracle-ee"
    CUSTOM_ORACLE_EE_CDB = "custom-oracle-ee-cdb"
    CUSTOM_SQLSERVER_EE = "custom-sqlserver-ee"
    CUSTOM_SQLSERVER_SE = "custom-sqlserver-se"
    CUSTOM_SQLSERVER_WEB = "custom-sqlserver-web"
    MARIADB = "mariadb"
    MYSQL = "mysql"
    NEPTUNE = "neptune"
    ORACLE_EE = "oracle-ee"
    ORACLE_EE_CDB = "oracle-ee-cdb"
    ORACLE_SE2 = "oracle-se2"
    ORACLE_SE2_CDB = "oracle-se2-cdb"
    POSTGRES = "postgres"
    SQLSERVER_EE = "sqlserver-ee"
    SQLSERVER_SE = "sqlserver-se"
    SQLSERVER_EX = "sqlserver-ex"
    SQLSERVER_WEB = "sqlserver-web"

    @classmethod
    def valid_db_instance_engine(self) -> List[str]:
        return sorted([item.value for item in DbInstanceEngine])


class ClusterEngine(str, Enum):
    AURORA_POSTGRESQL = "aurora-postgresql"
    AURORA_MYSQL = "aurora-mysql"
    NEPTUNE = "neptune"
    RDS_POSTGRESQL = "postgres"
    RDS_MYSQL = "mysql"

    @classmethod
    def list_cluster_engines(self) -> List[str]:
        return sorted([item.value for item in ClusterEngine])

    @classmethod
    def serverless_engines(self) -> List[str]:
        return [ClusterEngine.AURORA_MYSQL, ClusterEngine.AURORA_POSTGRESQL]


def get_object_value(obj: Any, attr: str) -> Any:
    """Retrieves an arbitrary attribute value from an object.

    Nested attributes can be specified using dot notation,
    e.g. 'parent.child'.

    :param object obj:
        A valid Python object.
    :param str attr:
        The attribute name of the value to retrieve from the object.
    :returns:
        The attribute value, if it exists, or None.
    :rtype:
        any
    """
    keys = attr.split(".")
    val = obj
    for key in keys:
        if hasattr(val, key):
            val = getattr(val, key)
        else:
            return None
    return val


def merge_filters(
    filters_to_update: Optional[Dict[str, Any]], filters_to_merge: Dict[str, Any]
) -> Dict[str, Any]:
    """Given two groups of filters, merge the second into the first.

    List values are appended instead of overwritten:

    >>> merge_filters({'filter-name': ['value1']}, {'filter-name':['value2']})
    >>> {'filter-name': ['value1', 'value2']}

    :param filters_to_update:
        The filters to update.
    :type filters_to_update:
        dict[str, list] or None
    :param filters_to_merge:
        The filters to merge.
    :type filters_to_merge:
        dict[str, list] or None
    :returns:
        The updated filters.
    :rtype:
        dict[str, list]
    """
    if filters_to_update is None:
        filters_to_update = {}
    if filters_to_merge is None:
        filters_to_merge = {}
    merge_dicts(filters_to_update, filters_to_merge, append_lists=True)
    return filters_to_update


def validate_filters(
    filters: Dict[str, Any], filter_defs: Dict[str, FilterDef]
) -> None:
    """Validates filters against a set of filter definitions.

    Raises standard Python exceptions which should be caught
    and translated to an appropriate AWS/Moto exception higher
    up the call stack.

    :param dict[str, list] filters:
        The filters to validate.
    :param dict[str, FilterDef] filter_defs:
        The filter definitions to validate against.
    :returns: None
    :rtype: None
    :raises KeyError:
        if filter name not found in the filter definitions.
    :raises ValueError:
        if filter values is an empty list.
    :raises NotImplementedError:
        if `moto` does not yet support this filter.
    """
    for filter_name, filter_values in filters.items():
        filter_def = filter_defs.get(filter_name)
        if filter_def is None:
            raise KeyError(f"Unrecognized filter name: {filter_name}")
        if not filter_values:
            raise ValueError(f"The list of {filter_def.description} must not be empty.")
        if filter_def.attrs_to_check is None:
            raise NotImplementedError(
                f"{filter_name} filter has not been implemented in Moto yet."
            )


def apply_filter(
    resources: Any, filters: Any, filter_defs: dict[str, FilterDef]
) -> Any:
    """Apply an arbitrary filter to a group of resources.

    :param dict[str, object] resources:
        A dictionary mapping resource identifiers to resource objects.
    :param dict[str, list] filters:
        The filters to apply.
    :param dict[str, FilterDef] filter_defs:
        The supported filter definitions for the resource type.
    :returns:
        The filtered collection of resources.
    :rtype:
        dict[str, object]
    """
    resources_filtered = OrderedDict()
    for identifier, obj in resources.items():
        matches_filter = False
        for filter_name, filter_values in filters.items():
            filter_def = filter_defs.get(filter_name)
            if filter_def is None or filter_def.attrs_to_check is None:
                continue
            for attr in filter_def.attrs_to_check:
                if filter_def.case_insensitive:
                    filter_values = list(map(lambda x: x.lower(), filter_values))
                if get_object_value(obj, attr) in filter_values:
                    matches_filter = True
                    break
            else:
                matches_filter = False
            if not matches_filter:
                break
        if matches_filter:
            resources_filtered[identifier] = obj
    return resources_filtered


def get_start_date_end_date(
    base_date: str, window: str
) -> Tuple[datetime.datetime, datetime.datetime]:
    """Gets the start date and end date given DDD:HH24:MM-DDD:HH24:MM.

    :param base_date:
        type datetime
    :param window:
        DDD:HH24:MM-DDD:HH24:MM
    :returns:
        Start and End Date in datetime format
    :rtype:
        tuple
    """
    days = {"mon": 1, "tue": 2, "wed": 3, "thu": 4, "fri": 5, "sat": 6, "sun": 7}
    start = datetime.datetime.strptime(
        base_date + " " + window[4:9], "%d/%m/%y %H:%M"
    ) + datetime.timedelta(days=days[window[0:3]])
    end = datetime.datetime.strptime(
        base_date + " " + window[14::], "%d/%m/%y %H:%M"
    ) + datetime.timedelta(days=days[window[10:13]])
    return start, end


def get_start_date_end_date_from_time(
    base_date: str, window: str
) -> Tuple[datetime.datetime, datetime.datetime, bool]:
    """Gets the start date and end date given HH24:MM-HH24:MM.

    :param window:
        HH24:MM-HH24:MM
    :returns:
        Start and End Date in datetime format
        along with flag for spills over a day
        This is useful when determine time overlaps
    :rtype:
        tuple
    """
    times = window.split("-")
    spillover = False
    start = datetime.datetime.strptime(base_date + " " + times[0], "%d/%m/%y %H:%M")
    end = datetime.datetime.strptime(base_date + " " + times[1], "%d/%m/%y %H:%M")
    if end < start:
        end += datetime.timedelta(days=1)
        spillover = True
    return start, end, spillover


def get_overlap_between_two_date_ranges(
    start_time_1: datetime.datetime,
    end_time_1: datetime.datetime,
    start_time_2: datetime.datetime,
    end_time_2: datetime.datetime,
) -> int:
    """
    Determines overlap between 2 date ranges. Returns the overlap in seconds.
    """
    latest_start = max(start_time_1, start_time_2)
    earliest_end = min(end_time_1, end_time_2)
    delta = earliest_end - latest_start
    return (delta.days * SECONDS_IN_ONE_DAY) + delta.seconds


def valid_preferred_maintenance_window(
    maintenance_window: Any, backup_window: Any
) -> Optional[str]:
    """Determines validity of preferred_maintenance_window

    :param maintenance_windown:
        type DDD:HH24:MM-DDD:HH24:MM
    :param backup_window:
        type HH24:MM-HH24:MM
    :returns:
        message
    :rtype:
        str
    """
    MINUTES_30 = 1800
    HOURS_24 = 86400
    base_date = datetime.datetime.now().strftime("%d/%m/%y")
    try:
        p = re.compile(
            "([a-z]{3}):([0-9]{2}):([0-9]{2})-([a-z]{3}):([0-9]{2}):([0-9]{2})"
        )
        if len(maintenance_window) != 19 or re.search(p, maintenance_window) is None:
            return f"Invalid maintenance window format: {maintenance_window}. Should be specified as a range ddd:hh24:mi-ddd:hh24:mi (24H Clock UTC). Example: Sun:23:45-Mon:00:15"
        if backup_window:
            (
                backup_window_start,
                backup_window_end,
                backup_spill,
            ) = get_start_date_end_date_from_time(base_date, backup_window)
            (
                maintenance_window_start,
                maintenance_window_end,
                maintenance_spill,
            ) = get_start_date_end_date_from_time(
                base_date, maintenance_window[4:10] + maintenance_window[14::]
            )
            if (
                get_overlap_between_two_date_ranges(
                    backup_window_start,
                    backup_window_end,
                    maintenance_window_start,
                    maintenance_window_end,
                )
                >= 0
            ):
                return "The backup window and maintenance window must not overlap."

            # Due to spill overs, adjust the windows
            elif maintenance_spill:
                backup_window_start += datetime.timedelta(days=1)
                backup_window_end += datetime.timedelta(days=1)
            elif backup_spill:
                maintenance_window_start += datetime.timedelta(days=1)
                maintenance_window_end += datetime.timedelta(days=1)

            # If spills, rerun overlap test with adjusted windows
            if maintenance_spill or backup_spill:
                if (
                    get_overlap_between_two_date_ranges(
                        backup_window_start,
                        backup_window_end,
                        maintenance_window_start,
                        maintenance_window_end,
                    )
                    >= 0
                ):
                    return "The backup window and maintenance window must not overlap."

        maintenance_window_start, maintenance_window_end = get_start_date_end_date(
            base_date, maintenance_window
        )
        delta = maintenance_window_end - maintenance_window_start
        delta_seconds = delta.seconds + (delta.days * SECONDS_IN_ONE_DAY)
        if delta_seconds >= MINUTES_30 and delta_seconds <= HOURS_24:
            return None
        elif delta_seconds >= 0 and delta_seconds <= MINUTES_30:
            return "The maintenance window must be at least 30 minutes."
        else:
            return "Maintenance window must be less than 24 hours."
    except Exception:
        return f"Invalid day:hour:minute value: {maintenance_window}"


ORDERABLE_DB_INSTANCE_ENCODING = {
    "Engine": "E",
    "EngineVersion": "EV",
    "DBInstanceClass": "DBIC",
    "LicenseModel": "L",
    "AvailabilityZones": "AZ",
    "MultiAZCapable": "MC",
    "ReadReplicaCapable": "RC",
    "Vpc": "V",
    "SupportsStorageEncryption": "SE",
    "StorageType": "ST",
    "SupportsIops": "SI",
    "SupportsEnhancedMonitoring": "SM",
    "SupportsIAMDatabaseAuthentication": "SIAM",
    "SupportsPerformanceInsights": "SPI",
    "AvailableProcessorFeatures": "APF",
    "SupportedEngineModes": "SEM",
    "SupportsKerberosAuthentication": "SK",
    "OutpostCapable": "O",
    "SupportedActivityStreamModes": "SSM",
    "SupportsGlobalDatabases": "SGD",
    "SupportsClusters": "SC",
    "SupportedNetworkTypes": "SN",
    "SupportsStorageThroughput": "SST",
}
ORDERABLE_DB_INSTANCE_DECODING = {
    v: k for (k, v) in ORDERABLE_DB_INSTANCE_ENCODING.items()
}


def encode_orderable_db_instance(db: Dict[str, Any]) -> Dict[str, Any]:
    encoded = copy.deepcopy(db)
    if "AvailabilityZones" in encoded:
        encoded["AvailabilityZones"] = [
            az["Name"] for az in encoded["AvailabilityZones"]
        ]
    return {
        ORDERABLE_DB_INSTANCE_ENCODING.get(key, key): value
        for key, value in encoded.items()
    }


def decode_orderable_db_instance(db: Dict[str, Any]) -> Dict[str, Any]:
    decoded = copy.deepcopy(db)
    decoded_az = ORDERABLE_DB_INSTANCE_ENCODING.get(
        "AvailabilityZones", "AvailabilityZones"
    )
    if decoded_az in decoded:
        decoded["AvailabilityZones"] = [{"Name": az} for az in decoded[decoded_az]]
    return {
        ORDERABLE_DB_INSTANCE_DECODING.get(key, key): value
        for key, value in decoded.items()
    }



import boto3
import pytest
from boto3 import Session
from botocore.client import ClientError

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID


@mock_aws
def test_get_public_access_block_for_account():
    client = boto3.client("s3control", region_name="us-west-2")

    # With an invalid account ID:
    with pytest.raises(ClientError) as ce_err:
        client.get_public_access_block(AccountId="111111111111")
    assert ce_err.value.response["Error"]["Code"] == "AccessDenied"

    # Without one defined:
    with pytest.raises(ClientError) as ce_err:
        client.get_public_access_block(AccountId=ACCOUNT_ID)
    assert (
        ce_err.value.response["Error"]["Code"] == "NoSuchPublicAccessBlockConfiguration"
    )

    # Put a with an invalid account ID:
    with pytest.raises(ClientError) as ce_err:
        client.put_public_access_block(
            AccountId="111111111111",
            PublicAccessBlockConfiguration={"BlockPublicAcls": True},
        )
    assert ce_err.value.response["Error"]["Code"] == "AccessDenied"

    # Put with an invalid PAB:
    with pytest.raises(ClientError) as ce_err:
        client.put_public_access_block(
            AccountId=ACCOUNT_ID, PublicAccessBlockConfiguration={}
        )
    assert ce_err.value.response["Error"]["Code"] == "InvalidRequest"
    assert (
        "Must specify at least one configuration."
        in ce_err.value.response["Error"]["Message"]
    )

    # Correct PAB:
    client.put_public_access_block(
        AccountId=ACCOUNT_ID,
        PublicAccessBlockConfiguration={
            "BlockPublicAcls": True,
            "IgnorePublicAcls": True,
            "BlockPublicPolicy": True,
            "RestrictPublicBuckets": True,
        },
    )

    # Get the correct PAB (for all regions):
    for region in Session().get_available_regions("s3control"):
        region_client = boto3.client("s3control", region_name=region)
        assert region_client.get_public_access_block(AccountId=ACCOUNT_ID)[
            "PublicAccessBlockConfiguration"
        ] == {
            "BlockPublicAcls": True,
            "IgnorePublicAcls": True,
            "BlockPublicPolicy": True,
            "RestrictPublicBuckets": True,
        }

    # Delete with an invalid account ID:
    with pytest.raises(ClientError) as ce_err:
        client.delete_public_access_block(AccountId="111111111111")
    assert ce_err.value.response["Error"]["Code"] == "AccessDenied"

    # Delete successfully:
    client.delete_public_access_block(AccountId=ACCOUNT_ID)

    # Confirm that it's deleted:
    with pytest.raises(ClientError) as ce_err:
        client.get_public_access_block(AccountId=ACCOUNT_ID)
    assert (
        ce_err.value.response["Error"]["Code"] == "NoSuchPublicAccessBlockConfiguration"
    )


@mock_aws
def test_storage_lens_configuration():
    client = boto3.client("s3control", region_name="us-east-2")
    config_id = "my-test-config-id"
    config = {
        "Id": "id-test",
        "AccountLevel": {
            "ActivityMetrics": {"IsEnabled": True},
            "BucketLevel": {
                "ActivityMetrics": {"IsEnabled": True},
                "PrefixLevel": {
                    "StorageMetrics": {
                        "IsEnabled": True,
                        "SelectionCriteria": {
                            "Delimiter": "string",
                            "MaxDepth": 123,
                            "MinStorageBytesPercentage": 100,
                        },
                    }
                },
                "AdvancedCostOptimizationMetrics": {"IsEnabled": True},
                "AdvancedDataProtectionMetrics": {"IsEnabled": True},
                "DetailedStatusCodesMetrics": {"IsEnabled": True},
            },
            "AdvancedCostOptimizationMetrics": {"IsEnabled": True},
            "AdvancedDataProtectionMetrics": {"IsEnabled": True},
            "DetailedStatusCodesMetrics": {"IsEnabled": True},
            "StorageLensGroupLevel": {
                "SelectionCriteria": {
                    "Include": [
                        "string",
                    ],
                    "Exclude": [
                        "string",
                    ],
                }
            },
        },
        "Include": {
            "Buckets": [
                "string",
            ],
            "Regions": [
                "string",
            ],
        },
        "Exclude": {
            "Buckets": [
                "string",
            ],
            "Regions": [
                "string",
            ],
        },
        "DataExport": {
            "S3BucketDestination": {
                "Format": "CSV",
                "OutputSchemaVersion": "V_1",
                "AccountId": ACCOUNT_ID,
                "Arn": "arn:aws:s3:::bucket_name",
                "Prefix": "/prefix",
                "Encryption": {"SSES3": {}},
            },
            "CloudWatchMetrics": {"IsEnabled": True},
        },
        "IsEnabled": True,
        "AwsOrg": {"Arn": "string"},
        "StorageLensArn": "string",
    }
    tags = [
        {
            "Key": "string",
            "Value": "string",
        },
    ]

    resp = client.put_storage_lens_configuration(
        AccountId=ACCOUNT_ID,
        ConfigId=config_id,
        StorageLensConfiguration=config,
        Tags=tags,
    )
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200

    # Get the configuration:
    resp = client.get_storage_lens_configuration(
        AccountId=ACCOUNT_ID, ConfigId=config_id
    )
    assert "StorageLensConfiguration" in resp
    assert resp["StorageLensConfiguration"]["Id"] == "id-test"
    s3_dest = resp["StorageLensConfiguration"]["DataExport"]["S3BucketDestination"]
    assert s3_dest["AccountId"] == ACCOUNT_ID
    assert s3_dest["Arn"] == "arn:aws:s3:::bucket_name"
    assert "Encryption" in s3_dest
    assert s3_dest["Encryption"]["SSES3"] == {}

    # List the configurations
    resp = client.list_storage_lens_configurations(AccountId=ACCOUNT_ID)
    assert len(resp["StorageLensConfigurationList"]) == 1
    # Note the ID is NOT the config ID from the put request
    # but the ID from the config itself
    assert resp["StorageLensConfigurationList"][0]["Id"] == "id-test"
    assert "StorageLensArn" in resp["StorageLensConfigurationList"][0]
    assert resp["StorageLensConfigurationList"][0]["IsEnabled"] is True



import random
import string
from typing import Any, Dict, Iterable, List


def random_id(uppercase: bool = True, length: int = 13) -> str:
    ascii_set = string.ascii_uppercase if uppercase else string.ascii_lowercase
    chars = list(range(10)) + list(ascii_set)
    resource_id = random.choice(ascii_set) + "".join(
        str(random.choice(chars)) for _ in range(length - 1)
    )
    return resource_id


def match_task_arn(task: Dict[str, Any], arns: List[str]) -> bool:
    return task["ReplicationTaskArn"] in arns


def match_task_id(task: Dict[str, Any], ids: List[str]) -> bool:
    return task["ReplicationTaskIdentifier"] in ids


def match_task_migration_type(task: Dict[str, Any], migration_types: List[str]) -> bool:
    return task["MigrationType"] in migration_types


def match_task_endpoint_arn(task: Dict[str, Any], endpoint_arns: List[str]) -> bool:
    return (
        task["SourceEndpointArn"] in endpoint_arns
        or task["TargetEndpointArn"] in endpoint_arns
    )


def match_task_replication_instance_arn(
    task: Dict[str, Any], replication_instance_arns: List[str]
) -> bool:
    return task["ReplicationInstanceArn"] in replication_instance_arns


task_filter_functions = {
    "replication-task-arn": match_task_arn,
    "replication-task-id": match_task_id,
    "migration-type": match_task_migration_type,
    "endpoint-arn": match_task_endpoint_arn,
    "replication-instance-arn": match_task_replication_instance_arn,
}


def filter_tasks(tasks: Iterable[Any], filters: List[Dict[str, Any]]) -> Any:
    matching_tasks = tasks

    for f in filters:
        filter_function = task_filter_functions.get(f["Name"])

        if not filter_function:
            continue

        matching_tasks = filter(
            lambda task: filter_function(task, f["Values"]), matching_tasks
        )

    return matching_tasks



from typing import Final


class _Missing:
    """A singleton class to represent a sentinel for a missing value."""

    def __bool__(self) -> bool:
        return False

    def __copy__(self):  # type: ignore[no-untyped-def]
        return self

    def __deepcopy__(self, _):  # type: ignore[no-untyped-def]
        return self

    def __repr__(self) -> str:
        return "<moto.MISSING>"


MISSING: Final = _Missing()



import re

from moto.ram.utils import (
    AWS_MANAGED_PERMISSIONS,
    RAM_RESOURCE_TYPES,
    format_ram_permission,
)


def test_format_ram_permission_defaults():
    name = "TestPermission"
    resource_type = "test:Resource"
    result = format_ram_permission(name, resource_type)
    assert result["name"] == name
    assert result["resourceType"] == resource_type
    assert result["version"] == "1"
    assert result["arn"] == f"arn:aws:ram::aws:permission/{name}"
    assert result["status"] == "ATTACHABLE"
    assert result["isResourceTypeDefault"] is True
    assert result["permissionType"] == "AWS_MANAGED"
    assert result["defaultVersion"] is True
    # creationTime and lastUpdatedTime should be a string in the expected format
    assert re.match(
        r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{3}", result["creationTime"]
    )
    assert result["creationTime"] == result["lastUpdatedTime"]


def test_format_ram_permission_custom_values():
    name = "CustomPermission"
    resource_type = "custom:Type"
    version = "2"
    arn_prefix = "arn:aws:ram::custom:permission/"
    status = "UNATTACHABLE"
    creation_time = "2023-01-01 12:00:00.000"
    last_updated_time = "2023-01-02 13:00:00.000"
    is_resource_type_default = False
    permission_type = "CUSTOM"
    default_version = False

    result = format_ram_permission(
        name=name,
        resource_type=resource_type,
        version=version,
        arn_prefix=arn_prefix,
        status=status,
        creation_time=creation_time,
        last_updated_time=last_updated_time,
        is_resource_type_default=is_resource_type_default,
        permission_type=permission_type,
        default_version=default_version,
    )
    assert result["name"] == name
    assert result["resourceType"] == resource_type
    assert result["version"] == version
    assert result["arn"] == f"{arn_prefix}{name}"
    assert result["status"] == status
    assert result["creationTime"] == creation_time
    assert result["lastUpdatedTime"] == last_updated_time
    assert result["isResourceTypeDefault"] is False
    assert result["permissionType"] == permission_type
    assert result["defaultVersion"] is False


def test_ram_resource_types_structure():
    for entry in RAM_RESOURCE_TYPES:
        assert "resourceType" in entry
        assert "serviceName" in entry
        assert "resourceRegionScope" in entry
        assert entry["resourceRegionScope"] in ("REGIONAL", "GLOBAL")


def test_aws_managed_permissions_structure():
    for perm in AWS_MANAGED_PERMISSIONS:
        assert "arn" in perm
        assert perm["arn"].startswith("arn:aws:ram::aws:permission/")
        assert "name" in perm
        assert "resourceType" in perm
        assert "creationTime" in perm
        assert "lastUpdatedTime" in perm
        assert "status" in perm
        assert "permissionType" in perm
        assert perm["permissionType"] == "AWS_MANAGED"
        assert isinstance(perm["isResourceTypeDefault"], bool)



"""NetworkFirewallBackend class with methods for supported APIs."""

from typing import Any, Dict, List, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.es.exceptions import ResourceNotFound
from moto.utilities.paginator import paginate
from moto.utilities.tagging_service import TaggingService

PAGINATION_MODEL = {
    "list_firewalls": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "unique_attribute": "arn",
    },
}


class NetworkFirewallModel(BaseModel):
    def __init__(
        self,
        account_id: str,
        region_name: str,
        firewall_name: str,
        firewall_policy_arn: str,
        vpc_id: str,
        subnet_mappings: List[str],
        delete_protection: bool,
        subnet_change_protection: bool,
        firewall_policy_change_protection: bool,
        description: str,
        tags: List[Dict[str, str]],
        encryption_configuration: Dict[str, str],
        enabled_analysis_types: List[str],
    ):
        self.firewall_name = firewall_name
        self.firewall_policy_arn = firewall_policy_arn
        self.vpc_id = vpc_id
        self.subnet_mappings = subnet_mappings
        self.delete_protection: bool = (
            delete_protection if delete_protection is not None else True
        )
        self.subnet_change_protection: bool = (
            subnet_change_protection if subnet_change_protection is not None else True
        )
        self.firewall_policy_change_protection: bool = (
            firewall_policy_change_protection
            if firewall_policy_change_protection is not None
            else True
        )
        self.description = description
        self.tags = tags
        self.encryption_configuration = encryption_configuration
        self.enabled_analysis_types = enabled_analysis_types

        self.arn = f"arn:aws:network-firewall:{region_name}:{account_id}:firewall/{self.firewall_name}"

        self.update_token = "1a2b3c4d-5e6f-7a8b-9c0d-1e2f3a4b5c6d"
        self.firewall_status = {
            "Status": "READY",
            "ConfigurationSyncStateSummary": "IN_SYNC",
        }
        self.logging_configs: Dict[str, List[Dict[str, Any]]] = {}

    def to_dict(self) -> Dict[str, Any]:
        return {
            "FirewallName": self.firewall_name,
            "FirewallArn": self.arn,
            "FirewallPolicyArn": self.firewall_policy_arn,
            "VpcId": self.vpc_id,
            "SubnetMappings": self.subnet_mappings,
            "DeleteProtection": self.delete_protection,
            "SubnetChangeProtection": self.subnet_change_protection,
            "FirewallPolicyChangeProtection": self.firewall_policy_change_protection,
            "Description": self.description,
            "Tags": self.tags,
            "EncryptionConfiguration": self.encryption_configuration,
            "EnabledAnalysisTypes": self.enabled_analysis_types,
        }


class NetworkFirewallBackend(BaseBackend):
    """Implementation of NetworkFirewall APIs."""

    def __init__(self, region_name: str, account_id: str) -> None:
        super().__init__(region_name, account_id)
        self.firewalls: Dict[str, NetworkFirewallModel] = {}
        self.tagger = TaggingService()

    def create_firewall(
        self,
        firewall_name: str,
        firewall_policy_arn: str,
        vpc_id: str,
        subnet_mappings: List[str],
        delete_protection: bool,
        subnet_change_protection: bool,
        firewall_policy_change_protection: bool,
        description: str,
        tags: List[Dict[str, str]],
        encryption_configuration: Dict[str, str],
        enabled_analysis_types: List[str],
    ) -> NetworkFirewallModel:
        firewall = NetworkFirewallModel(
            self.account_id,
            self.region_name,
            firewall_name=firewall_name,
            firewall_policy_arn=firewall_policy_arn,
            vpc_id=vpc_id,
            subnet_mappings=subnet_mappings,
            delete_protection=delete_protection,
            subnet_change_protection=subnet_change_protection,
            firewall_policy_change_protection=firewall_policy_change_protection,
            description=description,
            tags=tags,
            encryption_configuration=encryption_configuration,
            enabled_analysis_types=enabled_analysis_types,
        )
        self.firewalls[firewall.arn] = firewall

        if tags:
            self.tagger.tag_resource(firewall.arn, tags)

        return firewall

    def _get_firewall(
        self, firewall_arn: Optional[str], firewall_name: Optional[str]
    ) -> NetworkFirewallModel:
        if firewall_arn:
            if firewall_arn in self.firewalls:
                return self.firewalls[firewall_arn]
        for firewall in self.firewalls.values():
            if firewall.firewall_name == firewall_name:
                return firewall
        raise ResourceNotFound("NetworkFirewall", str(firewall_arn or firewall_name))

    def describe_logging_configuration(
        self, firewall_arn: str, firewall_name: str
    ) -> NetworkFirewallModel:
        firewall: NetworkFirewallModel = self._get_firewall(firewall_arn, firewall_name)
        return firewall

    def update_logging_configuration(
        self,
        firewall_arn: str,
        firewall_name: str,
        logging_configuration: Dict[str, List[Dict[str, Any]]],
    ) -> NetworkFirewallModel:
        firewall: NetworkFirewallModel = self._get_firewall(firewall_arn, firewall_name)
        firewall.logging_configs = logging_configuration
        return firewall

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_firewalls(self, vpc_ids: List[str]) -> List[NetworkFirewallModel]:
        firewalls = list(self.firewalls.values())
        if vpc_ids:
            firewalls = [fw for fw in firewalls if fw.vpc_id in vpc_ids]
        return firewalls

    def describe_firewall(
        self, firewall_name: str, firewall_arn: str
    ) -> NetworkFirewallModel:
        return self._get_firewall(firewall_arn, firewall_name)


networkfirewall_backends = BackendDict(NetworkFirewallBackend, "network-firewall")



from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Event, Thread
from typing import Optional, Tuple


class SimpleServer:
    def __init__(self, handler: type[BaseHTTPRequestHandler]) -> None:
        self._port = 0
        self._handler = handler

        self._thread: Optional[Thread] = None
        self._ip_address = "0.0.0.0"
        self._server: Optional[HTTPServer] = None
        self._server_ready_event = Event()

    def _server_entry(self) -> None:
        self._server = HTTPServer(("0.0.0.0", 0), self._handler)
        self._server_ready_event.set()
        self._server.serve_forever()

    def start(self) -> None:
        self._thread = Thread(target=self._server_entry, daemon=True)
        self._thread.start()
        self._server_ready_event.wait()

    def get_host_and_port(self) -> Tuple[str, int]:
        assert self._server
        host, port = self._server.server_address
        return str(host), port

    def stop(self) -> None:
        if self._server:
            self._server.shutdown()
        if self._thread:
            self._thread.join()



"""Handles incoming connectcampaigns requests, invokes methods, returns responses."""

import json
from typing import Dict, List, Optional, Tuple

from moto.core.responses import BaseResponse

from .models import ConnectCampaignServiceBackend, connectcampaigns_backends


class ConnectCampaignServiceResponse(BaseResponse):
    """Handler for ConnectCampaignService requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="connectcampaigns")

    @property
    def connectcampaigns_backend(self) -> ConnectCampaignServiceBackend:
        """Return backend instance specific for this region."""
        return connectcampaigns_backends[self.current_account][self.region]

    def create_campaign(self) -> str:
        params = json.loads(self.body)
        name = params.get("name")
        connect_instance_id = params.get("connectInstanceId")
        dialer_config = params.get("dialerConfig")
        outbound_call_config = params.get("outboundCallConfig")
        tags = params.get("tags", {})

        id, arn, tags = self.connectcampaigns_backend.create_campaign(
            name=name,
            connect_instance_id=connect_instance_id,
            dialer_config=dialer_config,
            outbound_call_config=outbound_call_config,
            tags=tags,
        )

        response = {"id": id, "arn": arn, "tags": tags}

        return json.dumps(response)

    def delete_campaign(self) -> str:
        id = self.path.split("/")[-1]

        self.connectcampaigns_backend.delete_campaign(
            id=id,
        )

        return "{}"

    def describe_campaign(self) -> str:
        id = self.path.split("/")[-1]

        campaign_details = self.connectcampaigns_backend.describe_campaign(
            id=id,
        )

        response = {"campaign": campaign_details}

        return json.dumps(response)

    def get_connect_instance_config(self) -> str:
        connect_instance_id = self.path.split("/")[-2]

        connect_instance_config = (
            self.connectcampaigns_backend.get_connect_instance_config(
                connect_instance_id=connect_instance_id,
            )
        )

        response = {"connectInstanceConfig": connect_instance_config}

        return json.dumps(response)

    def start_instance_onboarding_job(self) -> str:
        connect_instance_id = self.path.split("/")[-2]

        params = json.loads(self.body) if self.body else {}
        encryption_config = params.get("encryptionConfig", {})

        job_status = self.connectcampaigns_backend.start_instance_onboarding_job(
            connect_instance_id=connect_instance_id,
            encryption_config=encryption_config,
        )

        response = {"connectInstanceOnboardingJobStatus": job_status}

        return json.dumps(response)

    def start_campaign(self) -> str:
        id = self.path.split("/")[2]

        self.connectcampaigns_backend.start_campaign(
            id=id,
        )

        return "{}"

    def stop_campaign(self) -> str:
        id = self.path.split("/")[2]

        self.connectcampaigns_backend.stop_campaign(
            id=id,
        )

        return "{}"

    def pause_campaign(self) -> str:
        id = self.path.split("/")[2]

        self.connectcampaigns_backend.pause_campaign(
            id=id,
        )

        return "{}"

    def resume_campaign(self) -> str:
        id = self.path.split("/")[2]

        self.connectcampaigns_backend.resume_campaign(
            id=id,
        )

        return "{}"

    def get_campaign_state(self) -> str:
        id = self.path.split("/")[2]

        campaign_state = self.connectcampaigns_backend.get_campaign_state(
            id=id,
        )

        response = {"state": campaign_state}

        return json.dumps(response)

    def list_campaigns(self) -> str:
        max_results = self._get_param("maxResults")
        next_token = self._get_param("nextToken")
        filters = self._get_param("filters", {})

        campaigns_result: Tuple[List[Dict[str, str]], Optional[str]] = (
            self.connectcampaigns_backend.list_campaigns(
                filters=filters,
                max_results=max_results,
                next_token=next_token,
            )
        )

        response = {
            "campaignSummaryList": campaigns_result[0],
            "nextToken": campaigns_result[1],
        }
        return json.dumps(response)

    def tag_resource(self) -> str:
        arn = self._get_param("arn")
        tags = self._get_param("tags")
        self.connectcampaigns_backend.tag_resource(
            arn=arn,
            tags=tags,
        )
        return json.dumps(dict())

    def untag_resource(self) -> str:
        arn = self._get_param("arn")
        tag_keys = self._get_param("tagKeys")
        self.connectcampaigns_backend.untag_resource(
            arn=arn,
            tag_keys=tag_keys,
        )
        return json.dumps(dict())

    def list_tags_for_resource(self) -> str:
        arn = self._get_param("arn")
        tags = self.connectcampaigns_backend.list_tags_for_resource(
            arn=arn,
        )
        return json.dumps(dict(tags=tags))



"""Unit tests for codedeploy-supported APIs."""

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_create_application():
    client = boto3.client("codedeploy", region_name="ap-southeast-1")
    for platform in ["Server", "Lambda", "ECS"]:
        name = f"test-application-{platform}"
        response = client.create_application(
            applicationName=name,
            computePlatform=platform,
            tags=[{"Key": "Name", "Value": "Test"}],
        )
        assert "applicationId" in response

        resp = client.get_application(applicationName=name)
        application = resp["application"]
        assert application["applicationId"] == response["applicationId"]
        assert application["applicationName"] == name
        assert application["computePlatform"] == platform
        assert "createTime" in application


@mock_aws
def test_create_application_existing():
    client = boto3.client("codedeploy", region_name="ap-southeast-1")

    response = client.create_application(
        applicationName="sample_app", computePlatform="Server"
    )
    assert "applicationId" in response

    with pytest.raises(ClientError) as exc:
        response = client.create_application(
            applicationName="sample_app", computePlatform="Server"
        )
    assert exc.value.response["Error"]["Code"] == "ApplicationAlreadyExistsException"


@mock_aws
def test_create_deployment_group_required():
    client = boto3.client("codedeploy", region_name="eu-west-1")
    application_name = "mytestapp"
    client.create_application(
        applicationName=application_name, computePlatform="Lambda"
    )

    with pytest.raises(ClientError) as exc:
        client.create_deployment(applicationName=application_name)
    assert exc.value.response["Error"]["Code"] == "DeploymentGroupNameRequiredException"


@mock_aws
def test_create_deployment_revision_s3():
    client = boto3.client("codedeploy", region_name="us-west-2")

    application_name = "mytestapp"
    compute_platform = "Server"
    deployment_group_name = "test-deployment-group"
    service_role_arn = "arn:aws:iam::123456789012:role/CodeDeployDemoRole"
    description = "Test deployment"

    # Create an application
    client.create_application(
        applicationName=application_name, computePlatform=compute_platform
    )

    # Create a deployment group
    client.create_deployment_group(
        applicationName=application_name,
        deploymentGroupName=deployment_group_name,
        serviceRoleArn=service_role_arn,
    )

    revision_S3 = {
        "revisionType": "S3",
        "s3Location": {
            "bucket": "my-bucket",
            "key": "my-key",
            "bundleType": "zip",
            "version": "1",
            "eTag": "my-etag",
        },
    }

    response = client.create_deployment(
        applicationName=application_name,
        deploymentGroupName=deployment_group_name,
        revision=revision_S3,
        description=description,
    )
    deployment_id = response["deploymentId"]

    # Get the deployment
    response = client.get_deployment(deploymentId=deployment_id)

    assert "deploymentInfo" in response
    assert response["deploymentInfo"]["applicationName"] == application_name
    assert response["deploymentInfo"]["revision"]["revisionType"] == "S3"
    assert (
        response["deploymentInfo"]["revision"]["s3Location"]
        == revision_S3["s3Location"]
    )
    assert response["deploymentInfo"]["status"] == "Created"
    assert "createTime" in response["deploymentInfo"]


@mock_aws
def test_create_deployment_revision_github():
    client = boto3.client("codedeploy", region_name="us-west-2")

    application_name = "mytestapp-2"
    compute_platform = "Server"
    deployment_group_name = "test-deployment-group"
    service_role_arn = "arn:aws:iam::123456789012:role/CodeDeployDemoRole"
    description = "Test deployment"

    # Create an application
    client.create_application(
        applicationName=application_name, computePlatform=compute_platform
    )

    # Create a deployment group
    client.create_deployment_group(
        applicationName=application_name,
        deploymentGroupName=deployment_group_name,
        serviceRoleArn=service_role_arn,
    )

    revision_github = {
        "revisionType": "GitHub",
        "gitHubLocation": {
            "repository": "MyGitHubUserName/CodeDeployGitHubDemo",
            "commitId": "1234567890abcdef1234567890abcdef12345678",
        },
    }

    response = client.create_deployment(
        applicationName=application_name,
        deploymentGroupName=deployment_group_name,
        revision=revision_github,
        description=description,
    )
    deployment_id = response["deploymentId"]

    # Get the deployment
    response = client.get_deployment(deploymentId=deployment_id)
    assert "deploymentInfo" in response
    assert response["deploymentInfo"]["applicationName"] == application_name
    assert response["deploymentInfo"]["revision"]["revisionType"] == "GitHub"
    assert (
        response["deploymentInfo"]["revision"]["gitHubLocation"]
        == revision_github["gitHubLocation"]
    )
    assert response["deploymentInfo"]["status"] == "Created"
    assert "createTime" in response["deploymentInfo"]


@mock_aws
def test_create_deployment_nonexistent_app():
    client = boto3.client("codedeploy", region_name="eu-west-1")
    with pytest.raises(ClientError) as exc:
        client.create_deployment(applicationName="nonexistent-app")
    assert exc.value.response["Error"]["Code"] == "ApplicationDoesNotExistException"


@mock_aws
def test_create_deployment_nonexistent_group():
    client = boto3.client("codedeploy", region_name="eu-west-1")
    application_name = "mytestapp"

    client.create_application(
        applicationName=application_name, computePlatform="Lambda"
    )

    with pytest.raises(ClientError) as exc:
        client.create_deployment(
            applicationName=application_name,
            deploymentGroupName="non-existent-group-name",
            revision={
                "revisionType": "S3",
                "s3Location": {
                    "bucket": "my-bucket",
                    "key": "my-key",
                    "bundleType": "zip",
                    "version": "1",
                    "eTag": "my-etag",
                },
            },
        )
    assert exc.value.response["Error"]["Code"] == "DeploymentGroupDoesNotExistException"


@mock_aws
def test_create_deployment_group():
    client = boto3.client("codedeploy", region_name="eu-west-1")
    application_name = "mytestapp"
    deployment_group_name = "mytestdeploymentgroup"
    service_role_arn = "arn:aws:iam::123456789012:role/CodeDeployDemoRole"

    client.create_application(
        applicationName=application_name, computePlatform="Lambda"
    )

    response = client.create_deployment_group(
        applicationName=application_name,
        deploymentGroupName=deployment_group_name,
        serviceRoleArn=service_role_arn,
    )

    assert "deploymentGroupId" in response


@mock_aws
def test_create_deployment_group_app_nonexistent():
    client = boto3.client("codedeploy", region_name="eu-west-1")
    deployment_group_name = "mytestdeploymentgroup"
    service_role_arn = "arn:aws:iam::123456789012:role/CodeDeployDemoRole"

    with pytest.raises(ClientError) as exc:
        client.create_deployment_group(
            applicationName="nonexistent-app-name",
            deploymentGroupName=deployment_group_name,
            serviceRoleArn=service_role_arn,
        )
    assert exc.value.response["Error"]["Code"] == "ApplicationDoesNotExistException"


@mock_aws
def test_create_deployment_group_existing():
    client = boto3.client("codedeploy", region_name="eu-west-1")
    application_name = "mytestapp"
    deployment_group_name = "mytestdeploymentgroup"
    service_role_arn = "arn:aws:iam::123456789012:role/CodeDeployDemoRole"

    client.create_application(
        applicationName=application_name, computePlatform="Lambda"
    )

    client.create_deployment_group(
        applicationName=application_name,
        deploymentGroupName=deployment_group_name,
        serviceRoleArn=service_role_arn,
    )

    with pytest.raises(ClientError) as exc:
        client.create_deployment_group(
            applicationName=application_name,
            deploymentGroupName=deployment_group_name,
            serviceRoleArn=service_role_arn,
        )
    assert (
        exc.value.response["Error"]["Code"] == "DeploymentGroupAlreadyExistsException"
    )


@mock_aws
def test_get_deployment():
    client = boto3.client("codedeploy", region_name="eu-west-1")
    application_name = "mytestapp"
    deployment_group_name = "mytestdeploymentgroup"
    service_role_arn = "arn:aws:iam::123456789012:role/CodeDeployDemoRole"

    client.create_application(
        applicationName=application_name, computePlatform="Lambda"
    )
    client.create_deployment_group(
        applicationName=application_name,
        deploymentGroupName=deployment_group_name,
        serviceRoleArn=service_role_arn,
    )
    revision_content = "test-content"
    revision_sha256 = "test-sha256"
    description = "Test deployment"
    deployment_response = client.create_deployment(
        applicationName=application_name,
        deploymentGroupName=deployment_group_name,
        description=description,
        revision={
            "revisionType": "String",
            "string": {"content": revision_content, "sha256": revision_sha256},
        },
    )
    deployment_id = deployment_response["deploymentId"]

    response = client.get_deployment(deploymentId=deployment_id)
    assert "deploymentInfo" in response
    assert response["deploymentInfo"]["deploymentId"] == deployment_id
    assert response["deploymentInfo"]["applicationName"] == application_name
    assert response["deploymentInfo"]["deploymentGroupName"] == deployment_group_name
    assert response["deploymentInfo"]["deploymentId"] == deployment_id
    assert response["deploymentInfo"]["description"] == description
    assert response["deploymentInfo"]["revision"]["revisionType"] == "String"
    assert (
        response["deploymentInfo"]["revision"]["string"]["content"] == revision_content
    )
    assert response["deploymentInfo"]["revision"]["string"]["sha256"] == revision_sha256
    assert response["deploymentInfo"]["status"] == "Created"
    assert "createTime" in response["deploymentInfo"]


@mock_aws
def test_get_deployment_nonexistent_deployment():
    client = boto3.client("codedeploy", region_name="eu-west-1")
    with pytest.raises(ClientError) as exc:
        client.get_deployment(deploymentId="nonexistent-deployment-id")
    assert exc.value.response["Error"]["Code"] == "DeploymentDoesNotExistException"


@mock_aws
def test_get_deployment_group():
    client = boto3.client("codedeploy", region_name="eu-west-1")
    application_name = "mytestapp"
    deployment_group_name = "mytestdeploymentgroup"
    service_role_arn = "arn:aws:iam::123456789012:role/CodeDeployDemoRole"

    client.create_application(
        applicationName=application_name, computePlatform="Lambda"
    )

    response = client.create_deployment_group(
        applicationName=application_name,
        deploymentGroupName=deployment_group_name,
        serviceRoleArn=service_role_arn,
    )

    deployment_group_id = response["deploymentGroupId"]

    response = client.get_deployment_group(
        applicationName=application_name, deploymentGroupName=deployment_group_name
    )
    assert "deploymentGroupInfo" in response
    deployment_group_info = response["deploymentGroupInfo"]

    # required fields
    assert deployment_group_info["applicationName"] == application_name
    assert deployment_group_info["deploymentGroupId"] == deployment_group_id
    assert deployment_group_info["deploymentGroupName"] == deployment_group_name
    assert deployment_group_info["serviceRoleArn"] == service_role_arn

    # TODO assert other fields to be added
    # assert "alarmConfiguration" in deployment_group_info


@mock_aws
def test_get_deployment_group_nonexistent_app():
    client = boto3.client("codedeploy", region_name="eu-west-1")
    with pytest.raises(ClientError) as exc:
        client.get_deployment_group(
            applicationName="nonexistent-app-name",
            deploymentGroupName="mytestdeploymentgroup",
        )
    assert exc.value.response["Error"]["Code"] == "ApplicationDoesNotExistException"


@mock_aws
def test_get_deployment_group_nonexistent_group():
    client = boto3.client("codedeploy", region_name="eu-west-1")
    application_name = "mytestapp"

    client.create_application(
        applicationName=application_name, computePlatform="Lambda"
    )

    with pytest.raises(ClientError) as exc:
        client.get_deployment_group(
            applicationName=application_name,
            deploymentGroupName="nonexistent-group-name",
        )
    assert exc.value.response["Error"]["Code"] == "DeploymentGroupDoesNotExistException"


@mock_aws
def test_batch_get_applications_invalid_name():
    client = boto3.client("codedeploy", region_name="us-east-2")
    with pytest.raises(ClientError) as exc:
        client.batch_get_applications(applicationNames=["app_does_not_exist"])
    assert exc.value.response["Error"]["Code"] == "ApplicationDoesNotExistException"


@mock_aws
def test_batch_get_applications():
    client = boto3.client("codedeploy", region_name="us-east-2")
    client.create_application(applicationName="sample_app1", computePlatform="Lambda")
    client.create_application(applicationName="sample_app2", computePlatform="Server")
    client.create_application(applicationName="sample_app3", computePlatform="ECS")

    resp = client.batch_get_applications(
        applicationNames=["sample_app1", "sample_app2", "sample_app3"]
    )

    assert len(resp["applicationsInfo"]) == 3


@mock_aws
def test_batch_get_deployments():
    client = boto3.client("codedeploy", region_name="us-east-2")

    deployment_ids = []
    # create 2 deployments
    for i in range(0, 2):
        application_name = f"sample_app{i}"
        deployment_group_name = f"sample_deployment_group{i}"
        service_role_arn = "arn:aws:iam::123456789012:role/CodeDeployDemoRole"
        client.create_application(
            applicationName=application_name, computePlatform="Server"
        )
        client.create_deployment_group(
            applicationName=application_name,
            deploymentGroupName=deployment_group_name,
            serviceRoleArn=service_role_arn,
        )

        resp = client.create_deployment(
            applicationName=application_name,
            deploymentGroupName=deployment_group_name,
            revision={
                "revisionType": "S3",
                "s3Location": {
                    "bucket": "my-bucket",
                    "key": "my-key",
                    "bundleType": "zip",
                    "version": "1",
                    "eTag": "my-etag",
                },
            },
        )

        deployment_ids.append(resp["deploymentId"])

    resp = client.batch_get_deployments(deploymentIds=deployment_ids)
    assert len(resp["deploymentsInfo"]) == len(deployment_ids)


@mock_aws
def test_list_applications():
    client = boto3.client("codedeploy", region_name="us-east-2")
    resp = client.list_applications()
    assert len(resp["applications"]) == 0

    client.create_application(applicationName="sample_app", computePlatform="Server")
    client.create_application(applicationName="sample_app2", computePlatform="Server")
    resp = client.list_applications()
    assert len(resp["applications"]) == 2


@mock_aws
def test_list_deployments():
    client = boto3.client("codedeploy", region_name="ap-southeast-1")
    application_name = "mytestapp"
    deployment_group_name = "mytestdeploymentgroup"
    deployment_group_name2 = "mytestdeploymentgroup2"
    service_role_arn = "arn:aws:iam::123456789012:role/CodeDeployDemoRole"

    resp = client.list_deployments()
    assert len(resp["deployments"]) == 0

    client.create_application(
        applicationName=application_name, computePlatform="Server"
    )
    client.create_deployment_group(
        applicationName=application_name,
        deploymentGroupName=deployment_group_name,
        serviceRoleArn=service_role_arn,
    )

    client.create_deployment_group(
        applicationName=application_name,
        deploymentGroupName=deployment_group_name2,
        serviceRoleArn=service_role_arn,
    )
    client.create_deployment(
        applicationName=application_name,
        deploymentGroupName=deployment_group_name,
        revision={
            "revisionType": "S3",
            "s3Location": {
                "bucket": "my-bucket",
                "key": "my-key",
                "bundleType": "zip",
                "version": "1",
                "eTag": "my-etag",
            },
        },
    )

    client.create_deployment(
        applicationName=application_name,
        deploymentGroupName=deployment_group_name2,
        revision={
            "revisionType": "S3",
            "s3Location": {
                "bucket": "my-bucket",
                "key": "my-key",
                "bundleType": "zip",
                "version": "1",
                "eTag": "my-etag",
            },
        },
    )

    resp = client.list_deployments()
    assert len(resp["deployments"]) == 2


@mock_aws
def test_list_deployments_group_required():
    client = boto3.client("codedeploy", region_name="ap-southeast-1")
    with pytest.raises(ClientError) as exc:
        client.list_deployments(applicationName="mytestapp")
    assert exc.value.response["Error"]["Code"] == "DeploymentGroupNameRequiredException"


@mock_aws
def test_list_deployments_app_required():
    client = boto3.client("codedeploy", region_name="ap-southeast-1")
    with pytest.raises(ClientError) as exc:
        client.list_deployments(deploymentGroupName="mygroupname")
    assert exc.value.response["Error"]["Code"] == "ApplicationNameRequiredException"


@mock_aws
def test_list_deployment_groups():
    client = boto3.client("codedeploy", region_name="ap-southeast-1")

    application_name = "mytestapp"
    deployment_group_name = "mytestdeploymentgroup"
    deployment_group_name2 = "mytestdeploymentgroup2"
    service_role_arn = "arn:aws:iam::123456789012:role/CodeDeployDemoRole"

    resp = client.list_deployment_groups(applicationName=application_name)
    assert len(resp["deploymentGroups"]) == 0

    client.create_application(
        applicationName=application_name, computePlatform="Server"
    )
    client.create_deployment_group(
        applicationName=application_name,
        deploymentGroupName=deployment_group_name,
        serviceRoleArn=service_role_arn,
    )

    client.create_deployment_group(
        applicationName=application_name,
        deploymentGroupName=deployment_group_name2,
        serviceRoleArn=service_role_arn,
    )

    resp = client.list_deployment_groups(applicationName=application_name)
    assert len(resp["deploymentGroups"]) == 2

    # ensure both deployment group names are in the response
    assert set([deployment_group_name, deployment_group_name2]) == set(
        resp["deploymentGroups"]
    )


@mock_aws
def test_application_tagging():
    client = boto3.client("codedeploy", region_name="us-west-2")
    app_name = "test-tag-application"
    initial_tags = [{"Key": "Environment", "Value": "Test"}]

    response = client.create_application(
        applicationName=app_name, computePlatform="Server", tags=initial_tags
    )

    app_arn = f"arn:aws:codedeploy:us-west-2:123456789012:application:{app_name}"

    response = client.list_tags_for_resource(ResourceArn=app_arn)
    assert "Tags" in response
    assert len(response["Tags"]) == 1
    assert response["Tags"][0]["Key"] == "Environment"
    assert response["Tags"][0]["Value"] == "Test"


@mock_aws
def test_deployment_group_tagging():
    client = boto3.client("codedeploy", region_name="us-west-2")
    app_name = "test-tag-dg-app"
    dg_name = "test-tag-deployment-group"
    service_role_arn = "arn:aws:iam::123456789012:role/CodeDeployDemoRole"

    client.create_application(applicationName=app_name, computePlatform="Server")

    initial_tags = [{"Key": "Team", "Value": "Platform"}]
    client.create_deployment_group(
        applicationName=app_name,
        deploymentGroupName=dg_name,
        serviceRoleArn=service_role_arn,
        tags=initial_tags,
    )

    dg_arn = f"arn:aws:codedeploy:us-west-2:123456789012:deploymentgroup:{app_name}/{dg_name}"

    response = client.list_tags_for_resource(ResourceArn=dg_arn)
    assert "Tags" in response
    assert len(response["Tags"]) == 1
    assert response["Tags"][0]["Key"] == "Team"
    assert response["Tags"][0]["Value"] == "Platform"


@mock_aws
def test_deployment_inherits_tags():
    """Test that deployments inherit tags from deployment groups."""
    client = boto3.client("codedeploy", region_name="us-west-2")
    app_name = "test-tag-inherit-app"
    dg_name = "test-tag-inherit-group"
    service_role_arn = "arn:aws:iam::123456789012:role/CodeDeployDemoRole"

    client.create_application(applicationName=app_name, computePlatform="Server")

    dg_tags = [{"Key": "Environment", "Value": "Production"}]
    client.create_deployment_group(
        applicationName=app_name,
        deploymentGroupName=dg_name,
        serviceRoleArn=service_role_arn,
        tags=dg_tags,
    )

    response = client.create_deployment(
        applicationName=app_name,
        deploymentGroupName=dg_name,
        revision={
            "revisionType": "S3",
            "s3Location": {
                "bucket": "my-bucket",
                "key": "my-key",
                "bundleType": "zip",
                "version": "1",
                "eTag": "my-etag",
            },
        },
    )
    deployment_id = response["deploymentId"]

    deployment_arn = (
        f"arn:aws:codedeploy:us-west-2:123456789012:deployment:{deployment_id}"
    )

    response = client.list_tags_for_resource(ResourceArn=deployment_arn)
    assert "Tags" in response
    assert len(response["Tags"]) == 1



import boto3
import pytest

from moto import mock_aws


@mock_aws
@pytest.mark.parametrize(
    "region,partition", [("eu-west-1", "aws"), ("cn-north-1", "aws-cn")]
)
def test_put_insight_rule(region, partition):
    cloudwatch = boto3.client("cloudwatch", region)

    rule_name = "MySampleInsightRule"
    log_group_name = "my-log-group"
    rule_body = f"""
    {{
      "Schema": "CloudWatchLogRule",
      "LogGroupNames": ["{log_group_name}"],
      "LogFormat": "JSON",
      "Filter": "[timestamp, request, statusCode = *]",
      "Contribution": {{
        "Keys": ["statusCode"],
        "ValueOf": "Count"
      }}
    }}
    """
    cloudwatch.put_insight_rule(
        RuleName=rule_name, RuleDefinition=rule_body, RuleState="ENABLED"
    )

    rules = cloudwatch.describe_insight_rules()["InsightRules"]
    assert len(rules) == 1
    rule = rules[0]
    assert rule["Name"] == "MySampleInsightRule"
    assert rule["State"] == "ENABLED"
    assert rule["Schema"] == '{"Name" : "CloudWatchLogRule", "Version" : 1}'
    assert rule["Definition"] == rule_body
    assert rule["ManagedRule"] is not True


@mock_aws
def test_delete_insight_rules():
    cloudwatch = boto3.client("cloudwatch", region_name="eu-central-1")

    rule_name = "MySampleInsightRule"
    log_group_name = "my-log-group"
    rule_body = f"""
    {{
      "Schema": "CloudWatchLogRule",
      "LogGroupNames": ["{log_group_name}"],
      "LogFormat": "JSON",
      "Filter": "[timestamp, request, statusCode = *]",
      "Contribution": {{
        "Keys": ["statusCode"],
        "ValueOf": "Count"
      }}
    }}
    """

    cloudwatch.put_insight_rule(
        RuleName=rule_name, RuleDefinition=rule_body, RuleState="ENABLED"
    )

    cloudwatch.put_insight_rule(
        RuleName=f"{rule_name}_Backup", RuleDefinition=rule_body, RuleState="ENABLED"
    )
    rules = cloudwatch.describe_insight_rules()["InsightRules"]
    assert len(rules) == 2

    cloudwatch.delete_insight_rules(RuleNames=[rule_name])
    rules = cloudwatch.describe_insight_rules()["InsightRules"]
    assert len(rules) == 1

    rule = rules[0]
    assert rule["Name"] == "MySampleInsightRule_Backup"


@mock_aws
def test_disable_insight_rules():
    cloudwatch = boto3.client("cloudwatch", region_name="eu-central-1")

    rule_name = "MySampleInsightRule"
    log_group_name = "my-log-group"
    rule_body = f"""
    {{
      "Schema": "CloudWatchLogRule",
      "LogGroupNames": ["{log_group_name}"],
      "LogFormat": "JSON",
      "Filter": "[timestamp, request, statusCode = *]",
      "Contribution": {{
        "Keys": ["statusCode"],
        "ValueOf": "Count"
      }}
    }}
    """

    cloudwatch.put_insight_rule(
        RuleName=rule_name, RuleDefinition=rule_body, RuleState="ENABLED"
    )

    cloudwatch.put_insight_rule(
        RuleName=f"{rule_name}_Backup", RuleDefinition=rule_body, RuleState="ENABLED"
    )

    cloudwatch.disable_insight_rules(RuleNames=[rule_name])

    rules = cloudwatch.describe_insight_rules()

    # Ensuring correct rule was disabled, other should remain enabled
    for r in rules["InsightRules"]:
        if r["Name"] == rule_name:
            assert r["State"] == "DISABLED"
        else:
            assert r["State"] == "ENABLED"


@mock_aws
def test_enable_insight_rules():
    cloudwatch = boto3.client("cloudwatch", region_name="eu-central-1")

    rule_name = "MySampleInsightRule"
    log_group_name = "my-log-group"
    rule_body = f"""
    {{
      "Schema": "CloudWatchLogRule",
      "LogGroupNames": ["{log_group_name}"],
      "LogFormat": "JSON",
      "Filter": "[timestamp, request, statusCode = *]",
      "Contribution": {{
        "Keys": ["statusCode"],
        "ValueOf": "Count"
      }}
    }}
    """

    cloudwatch.put_insight_rule(
        RuleName=rule_name, RuleDefinition=rule_body, RuleState="DISABLED"
    )

    cloudwatch.put_insight_rule(
        RuleName=f"{rule_name}_Backup", RuleDefinition=rule_body, RuleState="DISABLED"
    )

    cloudwatch.enable_insight_rules(RuleNames=[rule_name])

    rules = cloudwatch.describe_insight_rules()

    # Ensuring correct rule was enabled, other should remain disabled
    for r in rules["InsightRules"]:
        if r["Name"] == rule_name:
            assert r["State"] == "ENABLED"
        else:
            assert r["State"] == "DISABLED"



# mypy: ignore-errors
import copy
import json
import os
from enum import Enum

import pytest
from botocore.model import OperationModel, ServiceModel

from moto.core.serialize import SERIALIZERS

TEST_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "protocols")
PROTOCOL_TEST_BLACKLIST = [
    "REST XML Event Stream",
    "RPC JSON Event Stream",
]


class TestType(Enum):
    __test__ = False  # Tell test runner to ignore this class

    INPUT = "input"
    OUTPUT = "output"


def _compliance_tests(test_type=None):
    inp = test_type is None or test_type is TestType.INPUT
    out = test_type is None or test_type is TestType.OUTPUT

    for full_path in _walk_files():
        if full_path.endswith(".json"):
            for model, case, protocol in _load_cases(full_path):
                if model.get("description") in PROTOCOL_TEST_BLACKLIST:
                    continue
                description = case["description"]
                test_name = f"{protocol}-protocol-{description}"
                if "params" in case and inp:
                    yield pytest.param(model, case, protocol, id=test_name)
                elif "response" in case and out:
                    yield pytest.param(model, case, protocol, id=test_name)


def _walk_files():
    for root, _, filenames in os.walk(TEST_DIR):
        for filename in filenames:
            yield os.path.join(root, filename)


def _load_cases(full_path):
    all_test_data = json.load(open(full_path))
    protocol = os.path.basename(full_path).split(".")[0]
    for test_data in all_test_data:
        cases = test_data.pop("cases")
        description = test_data["description"]
        for index, case in enumerate(cases):
            case["description"] = description
            case["id"] = index
            yield test_data, case, protocol


@pytest.mark.parametrize(
    "json_description, case, protocol", _compliance_tests(TestType.OUTPUT)
)
def test_output_compliance(json_description: dict, case: dict, protocol):
    service_description = copy.deepcopy(json_description)
    model = ServiceModel(service_description)
    operation_model = OperationModel(case["given"], model)
    protocol_serializer = SERIALIZERS[protocol]
    serializer = protocol_serializer(operation_model)
    result = case["result"] if "error" not in case else _create_exception(case)
    resp = serializer.serialize(result)  # _to_response(result)
    assert resp["body"] == case["response"]["body"]
    assert "Content-Type" in resp["headers"]
    protocol_to_content_type = {
        "ec2": "text/xml",
        "json": "application/x-amz-json-1.0",
        "query": "text/xml",
        "query-json": "application/json",
        "rest-xml": "text/xml",
        "rest-json": "application/json",
    }
    assert resp["headers"]["Content-Type"] == protocol_to_content_type[protocol]
    headers_expected = case["response"]["headers"]
    # TODO: Get rid of this if once we get the headers sorted for all responses
    if headers_expected:
        del resp["headers"]["Content-Type"]
        assert resp["headers"] == headers_expected
    assert resp["status_code"] == case["response"]["status_code"]


def _create_exception(case):
    exc = Exception()
    exc.code = case["errorCode"]
    if "errorMessage" in case:
        exc.message = case["errorMessage"]
        exc.Message = case["errorMessage"]
    for key, value in case["error"].items():
        setattr(exc, key, value)
    return exc



"""CodeDeployBackend class with methods for supported APIs."""

import uuid
from enum import Enum
from typing import Any, Dict, List, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import iso_8601_datetime_with_milliseconds
from moto.utilities.tagging_service import TaggingService

from .exceptions import (
    ApplicationAlreadyExistsException,
    ApplicationDoesNotExistException,
    ApplicationNameRequiredException,
    DeploymentDoesNotExistException,
    DeploymentGroupAlreadyExistsException,
    DeploymentGroupDoesNotExistException,
    DeploymentGroupNameRequiredException,
)


class Application(BaseModel):
    def __init__(
        self, application_name: str, compute_platform: str, tags: List[Dict[str, str]]
    ):
        self.id = str(uuid.uuid4())
        self.application_name = application_name
        self.compute_platform = compute_platform
        self.tags = tags.copy() if tags else []

        # Boto docs mention that the field should be datetime, but AWS API says number
        self.create_time = iso_8601_datetime_with_milliseconds()

        # these GitHub fields need to be set by the user in the console
        # so will be omitting them for now since they are not required and require console
        # self.github_account_name = ""
        # self.linked_to_github = False

    def to_dict(self) -> Dict[str, Any]:
        return {
            "applicationId": self.id,
            "applicationName": self.application_name,
            "createTime": self.create_time,
            "computePlatform": self.compute_platform,
        }


class CodeDeployDefault(str, Enum):
    # https://docs.aws.amazon.com/codedeploy/latest/userguide/deployment-configurations.html
    AllAtOnce = "AllAtOnce"
    HalfAtATime = "HalfAtATime"
    OneAtATime = "OneAtATime"

    def __str__(self) -> str:
        return f"{self.__class__.__name__}.{self.value}"


class AlarmConfiguration(BaseModel):
    def __init__(
        self,
        alarms: Optional[List[Dict[str, Any]]] = None,
        enabled: Optional[bool] = False,
        ignore_poll_alarm_failure: bool = False,
    ):
        self.alarms = alarms or []
        self.enabled = enabled
        self.ignore_poll_alarm_failure = ignore_poll_alarm_failure


class DeploymentGroup(BaseModel):
    def __init__(
        self,
        application: Application,
        deployment_group_name: str,
        deployment_config_name: Optional[str],
        ec2_tag_filters: Optional[List[Any]],
        on_premises_instance_tag_filters: Optional[List[Any]],
        auto_scaling_groups: Optional[List[str]],
        service_role_arn: str,
        trigger_configurations: Optional[List[Any]],
        alarm_configuration: Optional[AlarmConfiguration],
        auto_rollback_configuration: Optional[Dict[str, Any]],
        outdated_instances_strategy: Optional[str],
        deployment_style: Optional[Any],
        blue_green_deployment_configuration: Optional[Any],
        load_balancer_info: Optional[Any],
        ec2_tag_set: Optional[Any],
        ecs_services: Optional[List[Any]],
        on_premises_tag_set: Optional[Any],
        tags: Optional[List[Dict[str, str]]],
        termination_hook_enabled: Optional[bool],
    ):
        self.application = application
        self.deployment_group_name = deployment_group_name
        self.deployment_config_name = deployment_config_name
        self.ec2_tag_filters = ec2_tag_filters or []
        self.on_premises_instance_tag_filters = on_premises_instance_tag_filters or []
        self.auto_scaling_groups = auto_scaling_groups or []
        self.service_role_arn = service_role_arn
        self.trigger_configurations = trigger_configurations or []
        self.alarm_configuration = alarm_configuration
        self.auto_rollback_configuration = auto_rollback_configuration or {}
        self.outdated_instances_strategy = outdated_instances_strategy
        self.deployment_style = deployment_style or {}
        self.blue_green_deployment_configuration = (
            blue_green_deployment_configuration or {}
        )
        self.load_balancer_info = load_balancer_info or {}
        self.ec2_tag_set = ec2_tag_set or {}
        self.ecs_services = ecs_services or []
        self.on_premises_tag_set = on_premises_tag_set or {}
        self.tags = tags or []
        self.termination_hook_enabled = termination_hook_enabled
        self.deployment_group_id = str(uuid.uuid4())

    def to_dict(self) -> Dict[str, Any]:
        return {
            "applicationName": self.application.application_name,
            "deploymentGroupId": self.deployment_group_id,
            "deploymentGroupName": self.deployment_group_name,
            "deploymentConfigName": str(self.deployment_config_name),
            "ec2TagFilters": self.ec2_tag_filters,
            "onPremisesInstanceTagFilters": self.on_premises_instance_tag_filters,
            "autoScalingGroups": self.auto_scaling_groups,
            "serviceRoleArn": self.service_role_arn,
            "targetRevision": {},  # TODO
            "triggerConfigurations": self.trigger_configurations,
            "alarmConfiguration": {},  # TODO
            "autoRollbackConfiguration": self.auto_rollback_configuration,
            "deploymentStyle": self.deployment_style,
            "outdatedInstancesStrategy": self.outdated_instances_strategy,
            "blueGreenDeploymentConfiguration": self.blue_green_deployment_configuration,
            "loadBalancerInfo": self.load_balancer_info,
            "lastSuccessfulDeployment": {},  # TODO
            "lastAttemptedDeployment": {},  # TODO
            "ec2TagSet": self.ec2_tag_set,
            "onPremisesTagSet": self.on_premises_tag_set,
            "computePlatform": self.application.compute_platform,
            "ecsServices": self.ecs_services,
            "terminationHookEnabled": self.termination_hook_enabled,
        }


class DeploymentInfo(BaseModel):
    def __init__(
        self,
        application: Application,
        deployment_group: DeploymentGroup,
        revision: str,
        deployment_config_name: Optional[str],
        description: Optional[str],
        ignore_application_stop_failures: Optional[bool],
        targetInstances: Optional[Dict[str, Any]],
        auto_rollback_configuration: Optional[Dict[str, Any]],
        update_outdated_instances_only: Optional[bool],
        file_exists_behavior: Optional[str],
        override_alarm_configuration: Optional[AlarmConfiguration],
        creator: Optional[str],
    ):
        self.application = application
        self.deployment_group = deployment_group
        self.deployment_id = str(uuid.uuid4())
        self.application_name = application.application_name
        self.deployment_group_name = deployment_group.deployment_group_name
        self.revision = revision
        self.status = "Created"

        # Boto docs mention that the time fields should be datetime, but AWS API says number
        self.create_time = iso_8601_datetime_with_milliseconds()
        self.start_time = None  # iso_8601_datetime_with_milliseconds()
        self.complete_time = None  # iso_8601_datetime_with_milliseconds()

        # summary of deployment status of the instances in the deployment
        self.deployment_overview = {
            "Pending": 0,
            "InProgress": 0,
            "Succeeded": 0,
            "Failed": 0,
            "Skipped": 0,
            "Ready": 0,
        }
        self.description = description

        # the means by which the deployment was created: {user, autoscaling, codeDeployRollback, CodeDeployAutoUpdate}
        self.creator = "user" if not creator else creator

        self.deployment_config_name = deployment_config_name

        self.ignore_application_stop_failures = ignore_application_stop_failures
        self.target_instances = targetInstances
        self.auto_rollback_configuration = auto_rollback_configuration
        self.update_outdated_instances_only = update_outdated_instances_only
        self.instance_termination_wait_time_started = False

        self.additional_deployment_status_info = ""

        self.file_exists_behavior = file_exists_behavior
        self.deployment_status_messages: list[str] = []
        self.external_id = ""
        self.related_deployments: Dict[str, Any] = {}
        self.override_alarm_configuration = override_alarm_configuration

    def to_dict(self) -> Dict[str, Any]:
        return {
            "applicationName": self.application_name,
            "deploymentGroupName": self.deployment_group_name,
            "deploymentConfigName": str(self.deployment_config_name),
            "deploymentId": self.deployment_id,
            "previousRevision": {},  # TODO
            "revision": self.revision,
            "status": self.status,
            "errorInformation": {},  # TODO
            "createTime": self.create_time,
            "startTime": self.start_time,
            "completeTime": self.complete_time,
            "deploymentOverview": self.deployment_overview,
            "description": self.description,
            "creator": self.creator,
            "ignoreApplicationStopFailures": self.ignore_application_stop_failures,
            "autoRollbackConfiguration": self.auto_rollback_configuration,
            "updateOutdatedInstancesOnly": self.update_outdated_instances_only,
            "rollbackInfo": {},  # TODO information about a deployment rollback
            "deploymentStyle": self.deployment_group.deployment_style,
            "targetInstances": self.target_instances,
            "instanceTerminationWaitTimeStarted": self.instance_termination_wait_time_started,  # TODO
            "blueGreenDeploymentConfiguration": self.deployment_group.blue_green_deployment_configuration,
            "loadBalancerInfo": self.deployment_group.load_balancer_info,
            "additionalDeploymentStatusInfo": self.additional_deployment_status_info,  # TODO
            "fileExistsBehavior": self.file_exists_behavior,
            "deploymentStatusMessages": self.deployment_status_messages,  # TODO
            "computePlatform": self.application.compute_platform,
            "externalId": self.external_id,
            "relatedDeployments": self.related_deployments,  # TODO
            "overrideAlarmConfiguration": self.override_alarm_configuration,
        }


class CodeDeployBackend(BaseBackend):
    """Implementation of CodeDeploy APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.applications: Dict[str, Application] = {}
        self.deployments: Dict[str, DeploymentInfo] = {}
        self.deployment_groups: Dict[str, Dict[str, DeploymentGroup]] = {}
        self.tagger = TaggingService()

    def get_application(self, application_name: str) -> Application:
        if application_name not in self.applications:
            raise ApplicationDoesNotExistException(
                f"The application {application_name} does not exist with the user or AWS account."
            )
        return self.applications[application_name]

    def batch_get_applications(self, application_names: List[str]) -> List[Application]:
        applications_info = []
        for app_name in application_names:
            app_info = self.get_application(app_name)
            applications_info.append(app_info)

        return applications_info

    def get_deployment(self, deployment_id: str) -> DeploymentInfo:
        if deployment_id not in self.deployments:
            raise DeploymentDoesNotExistException(
                f"The deployment {deployment_id} does not exist with the user or AWS account."
            )
        return self.deployments[deployment_id]

    def get_deployment_group(
        self, application_name: str, deployment_group_name: str
    ) -> DeploymentGroup:
        if application_name not in self.applications:
            raise ApplicationDoesNotExistException(
                f"The application {application_name} does not exist with the user or AWS account."
            )

        # application can also exist but just not associated with a deployment group
        if (
            application_name not in self.deployment_groups
            or deployment_group_name not in self.deployment_groups[application_name]
        ):
            raise DeploymentGroupDoesNotExistException(
                f"The deployment group {deployment_group_name} does not exist with the user or AWS account."
            )
        return self.deployment_groups[application_name][deployment_group_name]

    def batch_get_deployments(self, deployment_ids: List[str]) -> List[DeploymentInfo]:
        deployments = []
        for id in deployment_ids:
            if id in self.deployments:
                deployment_info = self.deployments[id]
                deployments.append(deployment_info)

        return deployments

    def create_application(
        self, application_name: str, compute_platform: str, tags: List[Dict[str, str]]
    ) -> str:
        if application_name in self.applications:
            raise ApplicationAlreadyExistsException(
                f"The application {application_name} already exists with the user or AWS account."
            )

        app = Application(application_name, compute_platform, tags)
        self.applications[app.application_name] = app

        if tags:
            app_arn = f"arn:aws:codedeploy:{self.region_name}:{self.account_id}:application:{application_name}"
            self.tagger.tag_resource(app_arn, tags)

        return app.id

    def create_deployment(
        self,
        application_name: str,
        deployment_group_name: str,
        revision: str,
        deployment_config_name: Optional[str] = None,
        description: Optional[str] = None,
        ignore_application_stop_failures: Optional[bool] = None,
        target_instances: Optional[Any] = None,
        auto_rollback_configuration: Optional[Any] = None,
        update_outdated_instances_only: Optional[bool] = None,
        file_exists_behavior: Optional[str] = None,
        override_alarm_configuration: Optional[Any] = None,
    ) -> str:
        if application_name not in self.applications:
            raise ApplicationDoesNotExistException(
                f"The application {application_name} does not exist with the user or AWS account."
            )

        # Deployment Group Name appears to be optional in create_deployment boto3 documents
        # but seems required in most cases, depending on the deployment type
        # https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-codedeploy-deploymentgroup.html
        # assume required for now

        if deployment_group_name:
            if deployment_group_name not in self.deployment_groups.get(
                application_name, {}
            ):
                raise DeploymentGroupDoesNotExistException(
                    "Deployment group name does not exist."
                )
        else:
            raise DeploymentGroupNameRequiredException(
                "Deployment group name is required."
            )

        if not deployment_config_name:
            # get the deployment from the deployment group if config name is not specified
            deployment_config_name = self.deployment_groups[application_name][
                deployment_group_name
            ].deployment_config_name

        deployment = DeploymentInfo(
            self.applications[application_name],
            self.deployment_groups[application_name][deployment_group_name],
            revision,
            deployment_config_name,
            description,
            ignore_application_stop_failures,
            target_instances,
            auto_rollback_configuration,
            update_outdated_instances_only,
            file_exists_behavior,
            override_alarm_configuration,
            "user",
        )

        self.deployments[deployment.deployment_id] = deployment

        deployment_arn = f"arn:aws:codedeploy:{self.region_name}:{self.account_id}:deployment:{deployment.deployment_id}"
        if self.deployment_groups[application_name][deployment_group_name].tags:
            self.tagger.tag_resource(
                deployment_arn,
                self.deployment_groups[application_name][deployment_group_name].tags,
            )

        return deployment.deployment_id

    # TODO support all optional fields
    def create_deployment_group(
        self,
        application_name: str,
        deployment_group_name: str,
        deployment_config_name: Optional[str],
        ec2_tag_filters: Optional[List[Dict[str, str]]],
        on_premises_instance_tag_filters: Optional[List[Dict[str, str]]],
        auto_scaling_groups: Optional[List[str]],
        service_role_arn: str,
        trigger_configurations: Optional[List[Dict[str, Any]]] = None,
        alarm_configuration: Optional[AlarmConfiguration] = None,
        auto_rollback_configuration: Optional[Dict[str, Any]] = None,
        outdated_instances_strategy: Optional[str] = None,
        deployment_style: Optional[Dict[str, str]] = None,
        blue_green_deployment_configuration: Optional[Dict[str, Any]] = None,
        load_balancer_info: Optional[Dict[str, Any]] = None,
        ec2_tag_set: Optional[Dict[str, Any]] = None,
        ecs_services: Optional[List[Dict[str, str]]] = None,
        on_premises_tag_set: Optional[Dict[str, Any]] = None,
        tags: Optional[List[Dict[str, str]]] = None,
        termination_hook_enabled: Optional[bool] = None,
    ) -> str:
        if application_name not in self.applications:
            raise ApplicationDoesNotExistException(
                f"The application {application_name} does not exist with the user or AWS account."
            )

        if deployment_group_name in self.deployment_groups.get(application_name, {}):
            raise DeploymentGroupAlreadyExistsException(
                f"Deployment group {deployment_group_name} already exists."
            )

        # if deployment_config_name is not specified, use the default
        if not deployment_config_name:
            deployment_config_name = CodeDeployDefault.OneAtATime

        dg = DeploymentGroup(
            self.applications[application_name],
            deployment_group_name,
            deployment_config_name,
            ec2_tag_filters,
            on_premises_instance_tag_filters,
            auto_scaling_groups,
            service_role_arn,
            trigger_configurations,
            alarm_configuration,
            auto_rollback_configuration,
            outdated_instances_strategy,
            deployment_style,
            blue_green_deployment_configuration,
            load_balancer_info,
            ec2_tag_set,
            ecs_services,
            on_premises_tag_set,
            tags,
            termination_hook_enabled,
        )

        if application_name not in self.deployment_groups:
            self.deployment_groups[application_name] = {}
        self.deployment_groups[application_name][dg.deployment_group_name] = dg

        if tags:
            dg_arn = f"arn:aws:codedeploy:{self.region_name}:{self.account_id}:deploymentgroup:{application_name}/{deployment_group_name}"
            self.tagger.tag_resource(dg_arn, tags)

        return dg.deployment_group_id

    # TODO: implement pagination
    def list_applications(self) -> List[str]:
        return list(self.applications.keys())

    # TODO: implement pagination and complete filtering
    def list_deployments(
        self,
        application_name: str,
        deployment_group_name: str,
        external_id: str,
        include_only_statuses: List[str],
        create_time_range: Dict[str, Any],
    ) -> List[str]:
        # Ensure if applicationName is specified, then deploymentGroupName must be specified.
        # If deploymentGroupName is specified, application must be specified else error.
        if application_name and not deployment_group_name:
            raise DeploymentGroupNameRequiredException(
                "If applicationName is specified, then deploymentGroupName must be specified."
            )

        if deployment_group_name and not application_name:
            raise ApplicationNameRequiredException(
                "If deploymentGroupName is specified, applicationName must be specified."
            )

        def matches_filters(deployment: DeploymentInfo) -> bool:
            if application_name and deployment.application_name != application_name:
                return False
            if deployment_group_name:
                if application_name not in self.deployment_groups:
                    return False
                if (
                    deployment_group_name
                    not in self.deployment_groups[application_name]
                ):
                    return False
                if deployment.deployment_group_name != deployment_group_name:
                    return False
                if (
                    include_only_statuses
                    and deployment.status not in include_only_statuses
                ):
                    return False
            return True

        return [
            deployment.deployment_id
            for deployment in self.deployments.values()
            if matches_filters(deployment)
        ]

    # TODO: implement pagination
    def list_deployment_groups(
        self, application_name: str, next_token: str
    ) -> List[str]:
        if application_name not in self.deployment_groups:
            return []

        return [
            deployment_group.deployment_group_name
            for deployment_group in self.deployment_groups[application_name].values()
        ]

    def list_tags_for_resource(
        self, resource_arn: str
    ) -> Dict[str, List[Dict[str, str]]]:
        return self.tagger.list_tags_for_resource(resource_arn)

    def tag_resource(
        self, resource_arn: str, tags: List[Dict[str, str]]
    ) -> Dict[str, Any]:
        self.tagger.tag_resource(resource_arn, tags)
        return {}

    def untag_resource(self, resource_arn: str, tag_keys: List[str]) -> Dict[str, Any]:
        self.tagger.untag_resource_using_names(resource_arn, tag_keys)
        return {}


codedeploy_backends = BackendDict(CodeDeployBackend, "codedeploy")



from http.server import BaseHTTPRequestHandler
from typing import Any
from unittest import SkipTest

import pytest
import requests

from moto import settings
from tests.test_core.utilities import SimpleServer

url = "http://motoapi.amazonaws.com/moto-api/proxy/passthrough"


class WebRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self) -> None:
        self.send_response(200)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        self.wfile.write(b"real response")


@pytest.fixture
def server() -> Any:  # type: ignore[misc]
    server = SimpleServer(WebRequestHandler)
    server.start()
    yield server
    server.stop()


def test_real_request_errors() -> None:
    if not settings.is_test_proxy_mode():
        raise SkipTest("Can only be tested in ProxyMode")

    http_proxy = settings.test_proxy_mode_endpoint()
    https_proxy = settings.test_proxy_mode_endpoint()
    proxies = {"http": http_proxy, "https": https_proxy}

    # Delete all to ensure we're starting with a clean slate
    requests.delete(url, proxies=proxies)

    resp = requests.get("http://httpbin.org/robots.txt", proxies=proxies)
    assert resp.status_code == 404
    assert resp.content == b"AWS Service not recognized or supported"


def test_configure_passedthrough_urls() -> None:
    if not settings.is_test_proxy_mode():
        raise SkipTest("Can only be tested in ProxyMode")

    http_proxy = settings.test_proxy_mode_endpoint()
    https_proxy = settings.test_proxy_mode_endpoint()
    proxies = {"http": http_proxy, "https": https_proxy}

    # Delete all to ensure we're starting with a clean slate
    requests.delete(url, proxies=proxies)

    target1 = "http://httpbin.org/robots.txt"
    target2 = "http://othersite.org/"
    target3 = "https://othersite.org/"
    resp = requests.post(url, json={"http_urls": [target1]}, proxies=proxies)
    assert resp.status_code == 201
    assert resp.json() == {"http_urls": [target1], "https_hosts": []}

    # We can configure multiple URL's
    resp = requests.post(url, json={"http_urls": [target2]}, proxies=proxies)
    assert target1 in resp.json()["http_urls"]
    assert target2 in resp.json()["http_urls"]

    # Duplicate URL's are ignored
    requests.post(url, json={"http_urls": [target1]}, proxies=proxies)

    # We can retrieve the data
    resp = requests.get(url, proxies=proxies)
    assert target1 in resp.json()["http_urls"]
    assert target2 in resp.json()["http_urls"]
    assert resp.json()["https_hosts"] == []

    # Set HTTPS HOST for good measure
    resp = requests.post(url, json={"https_hosts": [target3]}, proxies=proxies)
    assert target1 in resp.json()["http_urls"]
    assert target2 in resp.json()["http_urls"]
    assert resp.json()["https_hosts"] == [target3]

    # We can delete all URL's in one go
    requests.delete(url, proxies=proxies)

    resp = requests.get(url, proxies=proxies)
    assert resp.json() == {"http_urls": [], "https_hosts": []}


def test_http_get_request_can_be_passed_through(server: Any) -> None:
    if not settings.is_test_proxy_mode():
        raise SkipTest("Can only be tested in ProxyMode")

    http_proxy = settings.test_proxy_mode_endpoint()
    https_proxy = settings.test_proxy_mode_endpoint()
    proxies = {"http": http_proxy, "https": https_proxy}

    # Delete all to ensure we're starting with a clean slate
    requests.delete(url, proxies=proxies)

    # Configure our URL as the one to passthrough
    server_url, port = server.get_host_and_port()
    target_url = f"http://127.0.0.1:{port}/robots.txt"
    requests.post(url, json={"http_urls": [target_url]}, proxies=proxies)

    resp = requests.get(target_url, proxies=proxies)
    assert resp.status_code == 200
    assert resp.content == b"real response"


def test_https_request_can_be_passed_through() -> None:
    raise SkipTest("Times out regularly")
    if not settings.is_test_proxy_mode():
        raise SkipTest("Can only be tested in ProxyMode")

    http_proxy = settings.test_proxy_mode_endpoint()
    https_proxy = settings.test_proxy_mode_endpoint()
    proxies = {"http": http_proxy, "https": https_proxy}

    # Delete all to ensure we're starting with a clean slate
    requests.delete(url, proxies=proxies)

    # Configure our URL as the one to passthrough
    target_url = "https://httpbin.org/ip"
    requests.post(url, json={"https_hosts": ["httpbin.org"]}, proxies=proxies)

    resp = requests.get(target_url, proxies=proxies)
    assert resp.status_code == 200
    assert "origin" in resp.json()
